/*****************************************************************************
 * I2C Routinen fr PIC30.
 * Copyright 2007 Andreas Kaiser
 *****************************************************************************/

#include "config.h"
#include "i2c.h"
#include "pic/i2c.h"
#include "util/rprintf.h"

#if 0
 #define DEBUG(f, ...)	rprintf(f, ## __VA_ARGS__)
#else
 #define DEBUG(f, ...)
#endif

//-----------------------------------------------------------------------------

static void
idle(void)
{
    while (I2CCON & 0x1F)
	;
}

static unsigned
send(unsigned data)
{
    idle();
    DEBUG("send %02X\n", data);
    I2CTRN = data;		// initiate transmission
    while (I2CSTATbits.TRSTAT)	// wait until completed
	;
    return I2CSTATbits.ACKSTAT;	// return acknowlege bit
}

static unsigned
recv(void)
{
    DEBUG("recv");
    I2CCONbits.RCEN = 1;	// initiate reception sequence
    while (!I2CSTATbits.RBF)	// wait until available
	;
    unsigned data = I2CRCV;
    DEBUG(" %02X\n", data);
    return data;
}

static void
ackn(unsigned ack)
{
    DEBUG("i2c:ackn %d\n", ack);
    if (ack)			// acknowledge bit value
	I2CCONbits.ACKDT = 0;
    else
	I2CCONbits.ACKDT = 1;
    I2CCONbits.RCEN = 1;	// initiate acknowledge sequence
    idle();
}

static unsigned
start(unsigned addr)
{
    DEBUG("start: %04X %04X\n", I2CCON, I2CSTAT);
    I2CCONbits.SEN = 1;		// set start condition
    return send(addr);
}

static unsigned
restart(unsigned addr)
{
    DEBUG("restart %02X\n", addr);
    I2CCONbits.RSEN = 1;	// set repeated start condition
    return send(addr);
}

static void
stop(void)
{
    I2CCONbits.PEN = 1;		// set stop condition
    idle();			// wait until completed
}

//-----------------------------------------------------------------------------

void
i2c_init(void)
{
    I2CCON = 0;						// default values
    I2CCONbits.SCLREL = 1;				// release SCL
    I2CCONbits.I2CEN = 1;				// enable module
    I2CBRG = (F_CPU / I2C_RATE) - (F_CPU / 1111111);	// -1 dropped due to cutoff rounding
    I2CMIEbit = 0;					// no master interrupts
    I2CSIEbit = 0;					// no slave  interrupts
}

unsigned
i2c_write(unsigned addr,
	  const uint8_t *ctrlBuf, unsigned ctrlLen,
	  const uint8_t *dataBuf, unsigned dataLen)
{
    unsigned i;

    if (start(addr)) {
	stop();
	return I2C_ADDR_NAK;
    }
    for (i = 0; i < ctrlLen; ++i) {
	if (send(ctrlBuf[i])) {
	    stop();
	    return I2C_DATA_NAK;
	}
    }
    for (i = 0; i < dataLen; ++i) {
	if (send(dataBuf[i])) {
	    stop();
	    return I2C_DATA_NAK;
	}
    }
    stop();
    return 0;
}

unsigned
i2c_read(unsigned addr,
	 const uint8_t *ctrlBuf, unsigned ctrlLen,
	       uint8_t *dataBuf, unsigned dataLen)
{
    unsigned i;

    if (ctrlLen) {
	if (start(addr)) {
	    stop();
	    return I2C_ADDR_NAK;
	}
	for (i = 0; i < ctrlLen; ++i) {
	    if (send(ctrlBuf[i])) {
		stop();
		return I2C_DATA_NAK;
	    }
	}
	if (restart(addr | 1)) {
	    stop();
	    return I2C_ADDR_NAK;
	}
    } else {
	if (start(addr | 1)) {
	    stop();
	    return I2C_ADDR_NAK;
	}
    }
    for (i = 0; i < dataLen; ++i) {
	dataBuf[i] = recv();
	ackn(i < dataLen-1);
    }
    stop();
    return 0;
}

